@@ -33,6 +33,7 @@
 #include <libgnome/gnome-i18n.h>
 #include <libgnome/gnome-url.h>
 #include <libgnome/gnome-program.h>
+#include <glade/glade.h>
 #include <string.h>
 #include "yelp-error.h"
 #include "yelp-html.h"
@@ -45,6 +46,8 @@
 #include "yelp-window.h"
 
 #define d(x)
+#define RESPONSE_PREV 1
+#define RESPONSE_NEXT 2
 
 typedef enum {
 	YELP_WINDOW_ACTION_BACK = 1,
@@ -92,6 +95,8 @@ static void        window_close_window_cb         (gpointer           data,
 						   GtkWidget         *widget);
 static void        window_find_cb                 (gpointer data,     guint section,
 						   GtkWidget         *widget);
+static void        window_find_again_cb           (gpointer data,     guint section,
+						   GtkWidget         *widget);
 static void        window_history_go_cb           (gpointer           data,
 						   guint              section,
 						   GtkWidget         *widget);
@@ -104,8 +109,13 @@ static void        window_go_index_cb             (gpointer           data,
 static void        window_about_cb                (gpointer           data,
 						   guint              section,
 						   GtkWidget         *widget);
+static gboolean    window_find_delete_event_cb    (GtkWidget         *widget,
+						   gpointer           user_data);
+static void        window_find_response_cb        (GtkWidget         *dialog ,
+						   gint               response,
+						   YelpWindow        *window);
+static YelpView *  window_get_active_view         (YelpWindow        *window);
 static GtkWidget * window_create_toolbar          (YelpWindow        *window);
-
 static GdkPixbuf * window_load_icon               (void);
 
 
@@ -134,7 +144,13 @@ struct _YelpWindowPriv {
 	YelpView       *index_view;
 	
 	GtkWidget      *find_dialog;
-
+	GtkWidget      *find_entry;
+	GtkWidget      *case_checkbutton;
+	GtkWidget      *wrap_checkbutton;
+	gchar          *find_string;
+	gboolean        match_case;
+	gboolean        wrap;
+	
 	YelpHistory    *history;
 
 	GtkItemFactory *item_factory;
@@ -145,14 +161,19 @@ struct _YelpWindowPriv {
 
 static GtkItemFactoryEntry menu_items[] = {
 	{N_("/_File"),              NULL,         0,                  0,                           "<Branch>"},
-	{N_("/File/_New window"),   NULL,         window_new_window_cb,   0,                           "<StockItem>", GTK_STOCK_NEW     },
-/*	{N_("/File/_Find in page..."), NULL,      window_find_cb,         0,                           "<StockItem>", GTK_STOCK_FIND   },*/
-	{N_("/File/_Close window"), NULL,         window_close_window_cb, 0,                           "<StockItem>", GTK_STOCK_CLOSE   },
+	{N_("/File/_New window"),   NULL,         window_new_window_cb,   0,                       "<StockItem>", GTK_STOCK_NEW     },
+	{N_("/File/_Close window"), NULL,         window_close_window_cb, 0,                       "<StockItem>", GTK_STOCK_CLOSE   },
+
+	{N_("/_Edit"),              NULL,         0,                  0,                           "<Branch>"},
+	{N_("/Edit/_Find in page..."), NULL,      window_find_cb,     0,                           "<StockItem>", GTK_STOCK_FIND    },
+	{N_("/Edit/_Find again"),   "<Control>g", window_find_again_cb, 0,                         "<StockItem>", GTK_STOCK_FIND    },
+	
 	{N_("/_Go"),                NULL,         0,                  0,                           "<Branch>"},
 	{N_("/Go/_Back"),           NULL,         window_history_go_cb,   YELP_WINDOW_ACTION_BACK,     "<StockItem>", GTK_STOCK_GO_BACK    },
 	{N_("/Go/_Forward"),        NULL,         window_history_go_cb,   YELP_WINDOW_ACTION_FORWARD,  "<StockItem>", GTK_STOCK_GO_FORWARD },
 	{N_("/Go/_Home"),           NULL,         window_go_home_cb,      0,                           "<StockItem>", GTK_STOCK_HOME },
 	{N_("/Go/_Index"),          NULL,         window_go_index_cb,     0,                           "<StockItem>", GTK_STOCK_INDEX },
+
 	{N_("/_Help"),              NULL,         0,                  0,                           "<Branch>"},
 	{N_("/Help/_About"),        NULL,         window_about_cb,        0,                           "<StockItem>", GNOME_STOCK_ABOUT },
 };
@@ -197,8 +218,11 @@ window_init (YelpWindow *window)
 	priv->content_view = NULL;
 	priv->index_view   = NULL;
 	
-	priv->history = yelp_history_new ();
+	priv->match_case   = FALSE;
+	priv->wrap         = TRUE;
 
+	priv->history = yelp_history_new ();
+	
 	g_signal_connect (priv->history, 
 			  "back_exists_changed",
 			  G_CALLBACK (window_toggle_history_back),
@@ -556,6 +580,49 @@ window_find_cb (gpointer data, guint section, GtkWidget *widget)
 {
 	YelpWindow     *window = data;
 	YelpWindowPriv *priv;
+	GladeXML       *glade;
+	
+	g_return_if_fail (YELP_IS_WINDOW (data));
+
+	window = YELP_WINDOW (data);
+
+	priv = window->priv;
+
+	if (!priv->find_dialog) {
+		glade = glade_xml_new (DATADIR "/yelp/ui/yelp.glade", "find_dialog", NULL);
+		if (!glade) {
+			g_warning ("Couldn't find necessary glade file " DATADIR "/yelp/ui/yelp.glade");
+			return;
+		}
+
+		priv->find_dialog = glade_xml_get_widget (glade, "find_dialog");
+
+		priv->find_entry = glade_xml_get_widget (glade, "find_entry");
+		priv->case_checkbutton = glade_xml_get_widget (glade, "case_check");
+		priv->wrap_checkbutton = glade_xml_get_widget (glade, "wrap_check");
+
+		g_signal_connect (priv->find_dialog,
+				  "delete_event",
+				  G_CALLBACK (window_find_delete_event_cb),
+				  NULL);
+
+		g_signal_connect (priv->find_dialog,
+				  "response",
+				  G_CALLBACK (window_find_response_cb),
+				  window);
+
+		g_object_unref (glade);
+	}
+
+	gtk_window_present (GTK_WINDOW (priv->find_dialog));
+}
+
+static void
+window_find_again_cb (gpointer data, guint section, GtkWidget *widget)
+{
+	YelpWindow     *window = data;
+	YelpWindowPriv *priv;
+	YelpView       *view;
 	YelpHtml       *html;
 	
 	g_return_if_fail (YELP_IS_WINDOW (data));
@@ -564,35 +631,18 @@ window_find_cb (gpointer data, guint section, GtkWidget *widget)
 
 	priv = window->priv;
 
-	/* To get rid of warning for now. */
-	if (0) {
-		window_find_cb (data, section, widget);
+	if (priv->find_string) {
+		view = window_get_active_view (window);
+		html = yelp_view_get_html (view);
+		
+		yelp_html_find (html,
+				priv->find_string,
+				priv->match_case,
+				priv->wrap,
+				TRUE);
 	}
-	
-	switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook))) {
-	case PAGE_TOC_VIEW:
-		html = yelp_view_get_html (priv->toc_view);
-		break;
-	case PAGE_CONTENT_VIEW:
-		html = yelp_view_get_html (priv->content_view);
-		break;
-	case PAGE_INDEX_VIEW:
-		html = yelp_view_get_html (priv->index_view);
-		break;
-	default:
-		g_assert_not_reached ();
-	}
-
-	if (priv->find_dialog) {
-		gtk_window_present (GTK_WINDOW (priv->find_dialog));
-	} else {
-		priv->find_dialog = NULL;
-	}
-
-	/* FIXME: Implement the dialog. */
-	yelp_html_find_next (html);
 }
-
+	
 static void
 window_history_go_cb (gpointer data, guint section, GtkWidget *widget)
 {
@@ -636,6 +686,87 @@ window_about_cb (gpointer data, guint section, GtkWidget *widget)
 	gtk_widget_show (about);
 }
 
+static gboolean
+window_find_delete_event_cb (GtkWidget *widget, gpointer user_data)
+{
+	gtk_widget_hide (widget);
+	
+	return TRUE;
+}
+
+static void
+window_find_response_cb (GtkWidget  *dialog ,
+			 gint        response,
+			 YelpWindow *window)
+{
+	YelpWindowPriv *priv;
+	YelpView       *view;
+	YelpHtml       *html;
+	const gchar    *tmp;
+	
+	priv = window->priv;
+
+	view = window_get_active_view (window);
+	html = yelp_view_get_html (view);
+
+	switch (response) {
+	case GTK_RESPONSE_CLOSE:
+		gtk_widget_hide (dialog);
+		break;
+
+	case RESPONSE_PREV:
+	case RESPONSE_NEXT:
+		tmp = gtk_entry_get_text (GTK_ENTRY (priv->find_entry));
+
+		priv->match_case = gtk_toggle_button_get_active (
+			GTK_TOGGLE_BUTTON (priv->case_checkbutton));
+
+		priv->wrap = gtk_toggle_button_get_active (
+			GTK_TOGGLE_BUTTON (priv->wrap_checkbutton));
+
+		g_free (priv->find_string);
+		
+		if (!priv->match_case) {
+			priv->find_string = g_utf8_casefold (tmp, -1);
+		} else {
+			priv->find_string = g_strdup (tmp);
+		}
+		
+		yelp_html_find (html,
+				priv->find_string,
+				priv->match_case,
+				priv->wrap,
+				response == RESPONSE_NEXT);
+		break;
+		
+	default:
+		break;
+	}
+	
+}
+
+static YelpView *
+window_get_active_view (YelpWindow *window)
+{
+	YelpWindowPriv *priv;
+	
+	priv = window->priv;
+	
+	switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (priv->notebook))) {
+	case PAGE_TOC_VIEW:
+		return priv->toc_view;
+	case PAGE_CONTENT_VIEW:
+		return priv->content_view;
+	case PAGE_INDEX_VIEW:
+		return priv->index_view;
+		break;
+	default:
+		g_assert_not_reached ();
+	}
+
+	return NULL;
+}
+
 static GtkWidget *
 window_create_toolbar (YelpWindow *window)
 {
